import logging
import os.path
import sys

import pandas as pd
import pyodbc
import validators

import inforion.ionapi.model.inforlogin as inforlogin
from inforion.helper.urlsplit import spliturl
from inforion.helper.util import get_db_credentials
from inforion.helper.validations import validate_db_credentials
from inforion.ionapi.controller import *
from inforion.ionapi.model import *
from inforion.merging.merging import merge_files
from inforion.transformation.transform import parallelize_tranformation

# from logger import get_logger

# from ionbasic import load_config

# Codee Junaid


def main_load(
    url=None,
    ionfile=None,
    program=None,
    method=None,
    dataframe=None,
    outputfile=None,
    start=None,
    end=None,
    on_progress=None,
):

    if validators.url(url) != True:
        logging.info("Error: URL is not valid")
        return "Error: URL is not valid"

    if os.path.exists(ionfile) == False:
        logging.info("Error: File does not exist")
        return "Error: File does not exist"
    else:
        inforlogin.load_config(ionfile)

    result = spliturl(url)

    if "Call" in result:
        if len(result["Call"]) > 0:
            if result["Call"] == "execute":
                inforlogin.load_config(ionfile)
                token = inforlogin.login()

                headers = inforlogin.header()
                if "Bearer" not in headers["Authorization"]:
                    return "Error: InforION Login is not working"
                if start is None or end is None:
                    return execute(
                        url,
                        headers,
                        program,
                        method,
                        dataframe,
                        outputfile,
                        on_progress,
                    )

                else:
                    return execute(
                        url,
                        headers,
                        program,
                        method,
                        dataframe,
                        outputfile,
                        start,
                        end,
                        on_progress,
                    )

            if result["Call"] == "executeSnd":

                config = inforlogin.load_config(ionfile)
                token = inforlogin.login()

                headers = inforlogin.header()
                if "Bearer" not in headers["Authorization"]:
                    return "InforION Login is not working"
                return executeSnd(
                    url, headers, program, method, dataframe, outputfile, start, end
                )

    if method == "checklogin":
        token = inforlogin.login()
        headers = inforlogin.header()
        return headers["Authorization"]


def main_transformation(
    mappingfile=None, mainsheet=None, stagingdata=None, outputfile=None
):

    if mappingfile is None:
        return "Error: Mapping file path missing"

    if os.path.exists(mappingfile) == False:
        return "Error: Mapping file does not exist"

    if mainsheet is None:
        return "Error: Main sheet name is empty"

    if stagingdata.empty:
        return "Error: Data frame is empty"

    try:
        return parallelize_tranformation(
            mappingfile, mainsheet, stagingdata, outputfile
        )
    except Exception as ex:
        if ex.message:
            return "There is an error while transforming the records. " + ex.message
        else:
            return "There is an unknown error while transforming the records."


def main_transformation_from_db(
    mappingfile=None, mainsheet=None, db_config=None, table_name=None, outputfile=None
):

    if mappingfile is None:
        return "Error: Mapping file path missing"

    if os.path.exists(mappingfile) == False:
        return "Error: Mapping file does not exist"

    if os.path.exists(db_config) == False:
        return "Error: DB Configurations does not exist."
    else:
        is_valid, code, message = validate_db_credentials(db_config)
        if not is_valid:
            return message

    if table_name is None:
        return "Error: Table name is missing."

    db_cred = get_db_credentials(db_config)
    conn_string = (
        "DRIVER={"
        + db_cred["driver"]
        + "};"
        + "SERVER={0};DATABASE={1};UID={2};PWD={3}".format(
            db_cred["servername"],
            db_cred["database"],
            db_cred["username"],
            db_cred["password"],
        )
    )
    sql_conn = pyodbc.connect(conn_string)

    query = "SELECT * FROM {0}".format(table_name)
    data = pd.read_sql(query, sql_conn)

    try:
        return parallelize_tranformation(mappingfile, mainsheet, data, outputfile)
    except Exception as ex:
        raise ex
        if hasattr(ex, "message"):
            return "There is an error while transforming the records. " + ex.message
        else:
            return "There is an unknown error while transforming the records."

        sys.exit(1)


def main_merge(
    mergesheet1=None,
    mergesheet2=None,
    mergeoutput=None,
    mergecol=None,
    mergetype="outer",
):

    if mergecol is None:
        return "Error: Merging column criteria not defined"

    if mergesheet1.empty:
        return "Error: First merge sheet frame is empty"

    if mergesheet2.empty:
        return "Error: Second merge sheet frame is empty"

    return merge_files(mergesheet1, mergesheet2, mergeoutput, mergecol, mergetype)



# Transformation of Staging Data into M3 format via mapping file
import datetime
import decimal
import logging
from functools import partial
from multiprocessing import Pool

import numpy as np
import pandas as pd

import inforion.transformation.transform_special as ts

# from logger import get_logger


def parallelize_tranformation(
    mappingfile, mainsheet, stagingdata, outputfile=None, n_cores=4
):

    # Read the file from given location
    xls = pd.ExcelFile(mappingfile)

    # to read all sheets to a map
    sheet_to_df_map = {}
    for sheet_name in xls.sheet_names:
        sheet_to_df_map[sheet_name] = xls.parse(sheet_name)

    # caching main sheet data
    main_cache = getMainSheetCache(sheet_to_df_map, mainsheet)
    # caching tab sheets data
    tabs_cache = getTabsMappingCache(sheet_to_df_map, main_cache)
    # getting a list of tabs containg wildcard characters
    wildcard_tabs = get_wild_card_tabs(tabs_cache)

    # Checking for special cases. e.g. Artikel-Lageort
    if mainsheet in ts.special_sheets:
        stagingdata = ts.handle_special_pre_transformations(
            sheet_to_df_map, mainsheet, stagingdata
        )

    # Transforming data in parallel
    df_split = np.array_split(stagingdata, n_cores)
    func = partial(
        transform_data,
        sheet_to_df_map,
        mainsheet,
        main_cache,
        tabs_cache,
        wildcard_tabs,
    )
    pool = Pool(n_cores)
    df = pd.concat(pool.map(func, df_split))
    pool.close()
    pool.join()

    # Saving output file
    if outputfile is not None:
        logging.info("Save to file: " + outputfile)
        writer = pd.ExcelWriter(outputfile, engine="xlsxwriter")
        df.to_excel(writer, sheet_name="Log Output", index=False)
        writer.save()
    return df


def getMainSheetCache(sheet_to_df_map, mainsheet):
    mapping_cache = []

    for index, row in sheet_to_df_map[mainsheet].iterrows():
        if index >= 9:
            row = row.replace(np.nan, "", regex=True)

            map = {}
            map["API_FIELD"] = row[15]

            if row[33] and not row[33] is np.nan:
                map["SOURCE"] = row[33]
            else:
                map["SOURCE"] = None

            map["FUNC_TYPE"] = row[36].strip().lower()

            map["FUNC_VAL"] = row[37]
            map["FUNC_ARG"] = row[38]

            mapping_cache.append(map)

    return mapping_cache


def getTabsMappingCache(sheet_to_df_map, mapping_cache):
    mapping_sheets_cache = {}

    for map in mapping_cache:
        if map["FUNC_TYPE"] == "tbl":
            index_val_col = 1  # index for the value column
            key_cols_count = 0  # no of key columns

            tab_key = map["FUNC_VAL"].strip()  # tab name for excel tab sheet
            tab_sub_keys = map["FUNC_ARG"].split("|")  # tab key columns information

            tab = {}  # tab containing all dictonaries for single excel tab sheet
            field_dict = {}  # dictionary for a single field

            for sub_val in tab_sub_keys:
                if "#:" in sub_val:
                    index_val_col = int(sub_val[2:])
                else:
                    key_cols_count = key_cols_count + 1

            index_val_col = key_cols_count + index_val_col - 1
            for i, val in sheet_to_df_map[tab_key].iterrows():
                multi_val_key = ""
                if i >= 7:
                    if str(val[0]) == "nan":
                        val[0] = ""
                    if key_cols_count > 1:
                        for index, sub_val in enumerate(tab_sub_keys):
                            if "#:" not in sub_val:
                                if multi_val_key == "":
                                    multi_val_key = str(val[index])
                                else:
                                    multi_val_key = (
                                        multi_val_key + "_" + str(val[index])
                                    )
                        field_dict[multi_val_key] = str(val[index_val_col])
                    else:
                        field_dict[str(val[0])] = str(val[int(index_val_col)])
            tab[str(map["API_FIELD"])] = field_dict
            if tab_key not in mapping_sheets_cache:
                mapping_sheets_cache[tab_key] = tab
            else:
                mapping_sheets_cache[tab_key].update(tab)

    return mapping_sheets_cache


def transform_data(
    _sheet_to_df_map, _mainsheet, sheet_cache, tabs_cache, wildcard_tabs, stagingdata
):
    rows_list = []

    for _, tb_row in stagingdata.iterrows():
        row_dict = {}
        for map in sheet_cache:
            if map["SOURCE"]:
                transform_source_column(
                    row_dict, map["API_FIELD"], tb_row, map["SOURCE"]
                )
            else:
                if map["FUNC_TYPE"] == "tbl":
                    tab = tabs_cache[map["FUNC_VAL"].strip()]
                    transform_sheet_table_mapping_column(
                        row_dict,
                        map["API_FIELD"],
                        tb_row,
                        tab,
                        map["FUNC_ARG"],
                        wildcard_tabs,
                    )
                elif map["FUNC_TYPE"] == "func":
                    transform_function_column(
                        row_dict,
                        map["API_FIELD"],
                        tb_row,
                        map["FUNC_VAL"],
                        map["FUNC_ARG"],
                    )
                elif map["FUNC_TYPE"] == "const":
                    transform_const_column(
                        row_dict, map["API_FIELD"], tb_row, map["FUNC_VAL"]
                    )

        rows_list.append(row_dict)

    df = pd.DataFrame(rows_list).replace("nan", "", regex=True)

    return df


def transform_source_column(map_row, map_col, tb_row, source_col):
    """Reads the value from source table against source col, and maps it to map_col in map_row"""
    source = clean(source_col)
    if source in tb_row:
        map_row[map_col] = str(tb_row[source])
    else:
        raise TransformationError(
            "Field '{}' mentioned in mapping sheet is not found.".format(source)
        )


def transform_sheet_table_mapping_column(
    map_row, map_col, tb_row, tab_sheet, tab_sheet_key, wildcard_tabs
):
    if tab_sheet_key and not tab_sheet_key is np.nan:
        db_key = ""  # actual key against which we lookup in sheet_tab

        # converting sheet_tab_key to db_key.
        sub_keys = tab_sheet_key.split("|")
        for sub_key in sub_keys:
            if "#:" not in sub_key:
                sub_key = clean(sub_key)
                if sub_key in tb_row:
                    if db_key == "":
                        db_key = str(tb_row[sub_key])
                    else:
                        db_key = db_key + "_" + str(tb_row[sub_key])
                else:
                    raise TransformationError(
                        "Field '{}' mentioned in mapping sheet is not found.".format(
                            sub_key
                        )
                    )

        if db_key in tab_sheet[map_col]:
            # assigning the values if key simply matches
            map_row[map_col] = str(tab_sheet[map_col][db_key])
        elif map_col in wildcard_tabs and wildcardcomparison_has_key(
            tab_sheet[map_col], db_key
        ):
            # checking if key matches with any wildcard entry in tab_sheet
            map_row[map_col] = wildcardcomparison_get_value(tab_sheet[map_col], db_key)
        elif "*" in tab_sheet[map_col]:
            # checking tab_sheet has '*' key, then we assign that values
            map_row[map_col] = str(tab_sheet[map_col]["*"])
        else:
            # if nothing matches, return the new converted key
            map_row[map_col] = str(db_key)


def transform_function_column(map_row, map_col, tb_row, func_name, func_args):
    """Reads the value from source table against source col, and maps it to map_col in map_row"""
    if func_name.strip().lower() == "div":
        data_values = func_args.split("|")
        with decimal.localcontext() as ctx:
            if data_values[2] != "":
                ctx.prec = int(data_values[2])
            result = decimal.Decimal(tb_row[clean(data_values[0])]) / decimal.Decimal(
                clean(data_values[1])
            )
    map_row[map_col] = result


def transform_const_column(map_row, map_col, tb_row, func_val):
    if isinstance(func_val, datetime.datetime):
        val = func_val.strftime("%Y%m%d")
        map_row[map_col] = str(val)
    else:
        map_row[map_col] = str(func_val)


def get_wild_card_tabs(tabs_cache):
    wildcard_tabs = []
    for key, tab_dict in tabs_cache.items():
        for key2, tab_dict2 in tab_dict.items():
            keys = list(
                filter(lambda x: len(x) > 1 and x.endswith("*"), tab_dict2.keys())
            )
            if len(keys) > 0:
                wildcard_tabs.append(key)
                break
    return wildcard_tabs


def wildcardcomparison_has_key(tab_dict, db_key):
    keys = list(filter(lambda x: len(x) > 1 and x.endswith("*"), tab_dict.keys()))
    for key in keys:
        skey = key[:-1]
        sdb_key = db_key[: len(skey)]
        if skey == sdb_key:
            return True
    return False


def wildcardcomparison_get_value(tab_dict, db_key):
    keys = list(filter(lambda x: len(x) > 1 and x.endswith("*"), tab_dict.keys()))
    for key in keys:
        skey = key[:-1]
        sdb_key = db_key[: len(skey)]
        if skey == sdb_key:
            return tab_dict[key]
    return db_key


def clean(string):
    str = string
    if str.startswith("["):
        str = str[1:]

    if str.endswith("]"):
        str = str[:-1]
    return str


class TransformationError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



# from logger import get_logger
import json
import logging as log

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests_toolbelt import MultipartEncoder

import inforion.ionapi.model.inforlogin as inforlogin


# logger = get_logger("my_module")


def get_messaging_ping():
    try:
        url = inforlogin.base_url() + "/IONSERVICES/api/ion/messaging/service/ping"
        headers = inforlogin.header()
        res = requests.get(url, headers=headers)
        log.info("messaging ping: {}".format(res.content))
        return res
    except Exception as e:
        log.error("Error ocurred " + str(e))


def post_messaging_v2_multipart_message(parameter_request, message_payload):
    try:
        url = (
            inforlogin.base_url()
            + "/IONSERVICES/api/ion/messaging/service/v2/multipartMessage"
        )
        data = MultipartEncoder(
            fields={
                "ParameterRequest": (
                    "filename",
                    json.dumps(parameter_request),
                    "application/json",
                ),
                "MessagePayload": (
                    "filename",
                    message_payload,
                    "application/octet-stream",
                ),
            }
        )
        headers = inforlogin.header()
        headers.update({"Content-Type": data.content_type})

        session = requests.Session()
        retries = Retry(total=10, backoff_factor=1, status_forcelist=[502, 503, 504])
        session.mount("https://", HTTPAdapter(max_retries=retries))
        res = session.post(url, headers=headers, data=data)
        log.info("messaging v2 multipart message: {}".format(res.content))
        return res
    except Exception as e:
        log.error("Error ocurred " + str(e))



import logging
import sys
from logging.handlers import TimedRotatingFileHandler

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
LOG_FILE = "inforion.log"


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name, add_log_file=False):
    logger = logging.getLogger(logger_name)
    # logger.setLevel(logging.INFO)
    logger.setLevel(logging.DEBUG)
    # logger.addHandler(get_console_handler())
    # requests_log = logging.getLogger("requests.packages.urllib3")
    # requests_log.setLevel(logging.DEBUG)
    # requests_log.propagate = True
    if add_log_file:
        logger.addHandler(get_file_handler())
    logger.propagate = False
    return logger



import json
import os
import shutil

import openpyxl
import xlsxwriter

import inforion.helper.api_fields_helper as afh

path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
apiDirPath = os.path.join(dir_path,'api-files')
templateFileName = dir_path + "/Mapping_Template.xlsx"

class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

def copyTemplateFile(outputfilename):
    filename = outputfilename
    if not outputfilename.endswith(".xlsx"):
        filename = filename + ".xlsx"
    return shutil.copy(templateFileName, filename)


def checkIfTemplateFileExists():
    return os.path.isfile(templateFileName)


def getTypeCode(type_str):
    if type_str == "string":
        return "A"
    elif type_str == "number":
        return "N"
    elif type_str == "date":
        return "D"


def updateSheetWithFieldsData(filename, fields_list):
    xfile = openpyxl.load_workbook(filename)

    # sheet = xfile.get_sheet_by_name('Mapping Template')
    sheet = xfile["Mapping Template"]

    current_row = 11
    added_fields = []
    for field in fields_list:
        # Serial Number

        if(field[2].upper() in  added_fields):
            continue

        sheet["A" + str(current_row)] = current_row - 10
        sheet["B" + str(current_row)] = field[0]

        sheet["P" + str(current_row)] = field[2]
        sheet["Q" + str(current_row)] = field[0]
        sheet["R" + str(current_row)] = field[1]
        sheet["S" + str(current_row)] = getTypeCode(field[4])
        sheet["V" + str(current_row)] = field[6]
        sheet["W" + str(current_row)] = 1 if field[5] else 0
        sheet["X" + str(current_row)] = field[3]

        added_fields.append(field[2].upper())
        current_row = current_row + 1

    xfile.save(filename)


def generate_api_template_file(program, outputfile):
    
    if not program:
        return "\033[91m" + "Error: Program name is missing" + "\033[0m"

    if not outputfile:
        return "\033[91m" + "Error: Output filename is missing" + "\033[0m"

    api_name = program
    filename = outputfile
    if not outputfile.endswith(".xlsx"):
        filename = filename + ".xlsx"

    fields = afh.get_fields_list_from_db(program)
    if(len(fields)>0):
        copyTemplateFile(filename)
        updateSheetWithFieldsData(filename, fields)
        print(bcolors.OKGREEN + "File generated successfully" + bcolors.ENDC)
        return True
    else:
        print(
            bcolors.FAIL
            + "Error: No fields information found for program '{0}'. Please make sure you have specified correct program name.".format(program)
            + bcolors.ENDC
        )
    return False


def getAPIFileDir():
    return apiDirPath



import json
import logging as log
from enum import Enum

import inforion.ionapi.model.inforlogin as inforlogin
import requests


class ObjectSchemaType(Enum):
    ANY = "ANY"
    BOD = "BOD"
    DSV = "DSV"
    JSON = "JSON"


def get_datacatalog_ping():
    url = inforlogin.base_url() + "/IONSERVICES/datacatalog/v1/status/ping"
    headers = inforlogin.header()
    res = requests.get(url, headers=headers)
    log.info("datacatalog ping: {}".format(res.content))
    return res


def delete_datacatalog_object(object_name):
    url = inforlogin.base_url() + "/IONSERVICES/datacatalog/v1/object/{}".format(
        object_name
    )
    headers = inforlogin.header()
    res = requests.delete(url, headers=headers)
    log.info("datacatalog delete: {}".format(res.content))
    return res


def post_datacatalog_object(
    object_name, object_type: ObjectSchemaType, schema, properties
):

    if object_type == ObjectSchemaType.ANY and (
        schema is not None or properties is not None
    ):
        raise ValueError("Schema and properties should be None")

    if (
        object_type == ObjectSchemaType.DSV or object_type == ObjectSchemaType.DSV
    ) and schema is None:
        raise ValueError("Schema cannot be None")

    url = inforlogin.base_url() + "/IONSERVICES/datacatalog/v1/object"
    headers = inforlogin.header()
    data = {
        "name": object_name,
        "type": object_type.value,
        "schema": schema,
        "properties": properties,
    }
    res = requests.post(url, headers=headers, data=json.dumps(data))
    log.info("datacatalog post: {}".format(res.content))
    return res



import xml.etree.ElementTree as et

import inforion.ln.soap_helper as shelper
import inforion.ln.xml_helper as xhelper
import pandas as pd
import requests


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


available_services = ["BusinessPartner_v3", "SalesOrder"]


def get_data(base_url, token, company, service_name):
    url = getRequestURL(base_url, service_name)

    request_body = shelper.get_list_request_xml(company, service_name)
    resp = shelper.execute_soap_request(url, token, request_body)

    data = getDataFromResponse(resp, service_name)
    return data


def export_data(url, token, company, service_name, outputfile):

    if service_name == "BusinessPartner":
        service_name == "BusinessPartner_v3"

    if service_name not in available_services:
        available_services_str = ", ".join(available_services)
        print(
            bcolors.FAIL
            + f"{service_name} is currently not available. Currently we only support {available_services_str}."
            + bcolors.ENDC
        )
        return

    data = get_data(url, token, company, service_name)
    if len(data) > 0:
        df_cols = xhelper.get_main_node_names(data[0])
        df = xhelper.get_data_frame_from_XML_Nodes_List(data, df_cols)
        df.to_excel(outputfile)
        print(
            bcolors.OKGREEN
            + "Exported data successfully to Excel file. "
            + bcolors.ENDC
        )
    else:
        print(bcolors.FAIL + f"No data found to export." + bcolors.ENDC)


def getDataFromResponse(resp, service_name):
    xroot = et.fromstring(resp.content)
    rows = []

    body = xroot[0]
    if isSuccess(body):
        listresp = body[0]
        listresp2 = listresp[0]
        data = listresp2[0]

        for partner in data.findall(service_name):
            rows.append(partner)
    else:
        errorMessage = getExceptionMessage(body)
        print(bcolors.FAIL + errorMessage + bcolors.ENDC)

    return rows


def getRequestURL(url, method):
    if url[-1] != "/":
        url = url + "/"

    if url.endswith(method):
        return url

    return url + method


def isSuccess(body):
    child = body[0]
    if "Fault" in child.tag:
        return False
    return True


def getExceptionMessage(body):
    child = body[0]
    detail = child[2]
    detailMessage = detail[1]
    return detailMessage.text



from inforion.helper.filehandling import *


def merge_files(mergesheet1, mergesheet2, outputfile, mergecol, mergetype):

    data = mergedata(mergesheet1, mergesheet2, mergecol, mergetype)

    savetodisk(outputfile, data)

    return True



import inforion as infor

import pandas as pd

dataframe = pd.read_excel("T-Vendor_Emailaddresse_CRS111_20200713.xlsx", dtype=str)

infor.main_load(
    "https://mingle-ionapi.eu1.inforcloudsuite.com/BVB_DEV/M3/m3api-rest/v2/execute",
    "FellowKey.ionapi",
    "CRS111MI",
    "Add,Change",
    dataframe,
    "T-Vendor_Emailaddresse_CRS111_20200713_report.xlsx",
    1,
    200,
)



from urllib.parse import urlparse


def url_change(url):
    if "ionapi" in url:
        url = "https://mingle-sso.eu1.inforcloudsuite.com/BVB_DEV/"
        return url
    else:
        return url


url = "https://mingle-ionapi.eu1.inforcloudsuite.com/BVB_DEV/"
if "ionapi" in url:
    path = urlparse(url).path

    result = "https://mingle-sso.eu1.inforcloudsuite.com" + path
    print("Original: ", url)
    print("Extracted: ", result)
